from tg import config, expose, request, require
from tg.controllers import TGController

from tgext.xmlrpc import XmlRpcController, xmlrpc
from repoze.what.predicates import Not, is_anonymous, has_permission

class TestRpcSubController(XmlRpcController):
    @xmlrpc([['string', 'array']], helpstr='joins an array of strings with spaces')
    def joinit(self, *p, **kw):
        return " ".join(p)
    
class TestRpcController(XmlRpcController):
    subrpc = TestRpcSubController()
    
    @xmlrpc([['int', 'array']], helpstr='sums an array of numbers')
    def addit(self, *p, **kw):
        return sum(p)
    
    @xmlrpc([['fault', 'none'], ['fault', 'int', 'string', 'struct', 'array']])
    def genfault(self, *p, **kw):
        1/0
        
    def sumit(self, *p, **kw):
        return sum(p)
    
class RootController(TGController):
    @expose()
    def index(self, *p, **kw):
        return 'hello world'

    xmlrpc = TestRpcController()
    


# -*- coding: utf-8 -*-
"""Controllers for the tgext.xmlrpc application."""

__all__ = ['XmlRpcController', 'xmlrpc']

import xmlrpclib
import urllib

from decorator import decorator

from pylons import config

from tg import request, response, expose
from tg.controllers import TGController
from tg.decorators import Decoration

class xmlrpc(object):
    def __init__(self, signatures, helpstr=""):
        self.signatures = signatures
        self.helpstr = helpstr
        
    def __call__(self, func):
        deco = decorator(self.wrap, func)
        exposed = Decoration.get_decoration(deco)
        deco.signatures = self.signatures
        deco.helpstr = self.helpstr
        exposed.register_template_engine(\
            'text/xml', config.get('default_renderer', ['mako'])[0],
            '', [])
        return deco
    
    def wrap(self, func, *p, **kw):
        try:
            try:
                parms, method = xmlrpclib.loads(request.body)
            except:
                parms, method = xmlrpclib.loads(urllib.unquote_plus(request.body))
            rpcresponse = xmlrpclib.dumps((func(p[0], *parms),), methodresponse=1)
        except xmlrpclib.Fault, fault:
            rpcresponse = xmlrpclib.dumps(fault)
        except Exception, e:
            rpcresponse = xmlrpclib.dumps(xmlrpclib.Fault(1, "%s:%s" % (str(type(e)), str(e))))
        return rpcresponse
        
class XmlRpcController(TGController):
    @expose()
    def index(self):
        header = '<html>\n<head>\n<title>XMLRPC Index Page</title>\n'
        style = '<style type="text/css">dt {font-weight:bold}</style>\n</head>\n</body>\n<dl>\n'
        footer = '</dl>\n</body>\n</html>\n'
        dlent = []
        for method in self._gather_all_methods('', self):
            dlent.append('<dt>%s</dt>\n<dd>\n<dl>\n' % (method))
            kall = self._find_method(method)
            dlent.append('<dt>Help String:</dt>\n<dd>%s</dd>\n' % (kall.helpstr))
            for sig in kall.signatures:
                retval = sig[:1]
                parms = sig[1:]
                dlent.append('<dt>Method Signature</dt>\n<dd>\n<dl>\n')
                dlent.append('<dt>Returns:</dt>\n<dd>%s</dd>\n' % (retval))
                dlent.append('<dt>Parameters:</dt>\n<dd>%s</dd>\n' % (parms))
                dlent.append('</dl>\n</dd>\n')
            dlent.append('</dl>\n')
        return "%s%s%s%s" % (header, style, ''.join(dlent), footer)
    
    @expose(content_type='text/xml')
    def rpcfault(self, *p, **kw):
        return xmlrpclib.dumps(xmlrpclib.Fault(1, p[0]))
    
    @expose(content_type='text/xml')
    def _system_methodHelp(self, *p, **kw):
        try:
            method = self._find_method(p[0])
            rpcresponse = xmlrpclib.dumps((method.helpstr,), methodresponse=1)
        except:
            rpcresponse = xmlrpclib.dumps(('Invalid method',), methodresponse=1)
        return rpcresponse
    
    @expose(content_type='text/xml')
    def _system_methodSignature(self, *p, **kw):
        try:
            method = self._find_method(p[0])
            rpcresponse = xmlrpclib.dumps((method.signatures,), methodresponse=1)
        except:
            rpcresponse = xmlrpclib.dumps(('Invalid method',), methodresponse=1)
        return rpcresponse
    
    @expose(content_type='text/xml')
    def _system_listMethods(self, *p, **kw):
        methods = self._gather_all_methods('', self)
        return xmlrpclib.dumps((methods,), methodresponse=1)
            
    def _gather_all_methods(self, prefix, controller):
        methods = []
        if prefix != '':
            prefix = prefix + '.'
        for attrname in dir(controller):
            attr = getattr(controller, attrname)
            if hasattr(attr, 'signatures'):
                methods.append('%s%s' % (prefix, attrname))
            if isinstance(attr, XmlRpcController):
                methods.extend(self._gather_all_methods(prefix + attrname, attr))
        return methods
        
    def _find_method(self, method):
        mvals = method.split('.')
        midx = 0
        controller = self
        result = None
        while not result and controller and midx < len(mvals):
            controller = getattr(controller, mvals[midx], None)
            if controller is None:
                break
            if midx == len(mvals)-1:
                result = controller
            midx = midx + 1
        return result
    
    def _dispatch(self, state, remainder, parms=None, method=None):
        if remainder:
            return self._dispatch_first_found_default_or_lookup(state, remainder)
        
        if request.method.lower() == 'get':
            state.add_method(self.index, [])
            return state
        
        if not parms and not method:
            try:
                parms, method = xmlrpclib.loads(request.body)
            except:
                try:
                    parms, method = xmlrpclib.loads(urllib.unquote_plus(request.body))
                except:
                    state.add_method(self.rpcfault, ['Unable to decode request body "||%s||"' % (request.body)])
                    return state
            
        if method.startswith('system.'):
            methodname = '_%s' % (method.replace('.', '_'))
            if not getattr(self, methodname, None):
                state.add_method(self.rpcfault, ['Invalid system method called: %s' % (method)])
            else:
                state.add_method(getattr(self, methodname), parms)
            return state

        mvals = method.split('.')
        if len(mvals) > 1:
            subcon = getattr(self, mvals[0], None)
            if subcon:
                if isinstance(subcon, XmlRpcController):
                    return subcon._dispatch(state, remainder, parms, "".join(mvals[1:]))
                else:
                    return subcon._dispatch_controller(mvals[0], self, state, mvals[1:])
            else:
                return self._dispatch_first_found_default_or_lookup(state, remainder)
        else:
            method = getattr(self, mvals[0], None)
            if method:
                if getattr(method, 'signatures', None) is not None:
                    state.add_method(method, [])
                    return state
                else:
                    state.add_method(self.rpcfault, ['Invalid XMLRPC Method'])
                    return state
        return self._dispatch_first_found_default_or_lookup(state, remainder)



# -*- coding: utf-8 -*-
try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

setup(
    name='tgext.xmlrpc',
    version='0.6',
    description='TurboGears XMLRPC Controller',
    author='Michael J. Pedersen',
    author_email='m.pedersen@icelus.org',
    url='http://bitbucker.org/pedersen/tgext.xmlrpc',
    install_requires=[
        "TurboGears2 >= 2.1b2",
        "decorator",
        ],
    setup_requires=["PasteScript >= 1.7"],
    paster_plugins=['PasteScript', 'Pylons', 'TurboGears2', 'tg.devtools'],
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['tgext'],
    include_package_data=True,
    test_suite='nose.collector',
    tests_require=['WebTest', 'BeautifulSoup'],
    message_extractors={'tgextxmlrpc': [
            ('**.py', 'python', None),
            ('templates/**.mako', 'mako', None),
            ('templates/**.html', 'genshi', None),
            ('public/**', 'ignore', None)]},
    entry_points="""
    
    """,
)



import os, sys
import xmlrpclib

from pylons import config
from tg.test_stack import TestConfig, app_from_config
from tg.util import Bunch

import tgext.xmlrpc.test

root = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, root)
paths=Bunch(
            root=root,
            controllers=os.path.join(root, 'controllers'),
            static_files=os.path.join(root, 'static'),
            templates=os.path.join(root, 'templates')
            )

base_config = TestConfig(folder = 'rendering',
                         values = {'use_sqlalchemy': False,
                                   'use_toscawidgets2': False,
                                   'full_stack': False,
                                   'pylons.helpers': Bunch(),
                                   'renderers': ['genshi'],
                                   'use_legacy_renderer': False,
                                   'default_renderer':'genshi',
                                   'use_dotted_templatenames': True,
                                   'paths':paths,
                                   'package':tgext.xmlrpc.test,
                                   'beaker.session.secret': 'ChAnGeMe',
                                   'beaker.session.key': 'tgext.xmlrpc.test',
                                  }
                         )

class TestXmlRpcController:
    def __init__(self):
        self.app = app_from_config(base_config)
  
    def test_tgbase(self):
        resp = self.app.get('/')
        assert 'hello world' in resp, resp
        
    def test_helppage(self):
        resp = self.app.get('/xmlrpc')
        assert 'Help String' in resp, resp
        assert 'Method Signature' in resp, resp
        assert 'Returns' in resp, resp
        assert 'Parameters' in resp, resp
        assert 'addit' in resp, resp
        assert 'genfault' in resp, resp
        assert 'subrpc.joinit' in resp, resp
    
    def test_xmlrpc_correct_argcount(self):
        resp = self.app.post('/xmlrpc', xmlrpclib.dumps((1,2), 'addit'))
        resp = xmlrpclib.loads(resp.body)
        assert resp[0][0] == 3, resp

    def test_subrpc(self):
        resp = self.app.post('/xmlrpc', xmlrpclib.dumps(("hello","world","i","mean","it"), 'subrpc.joinit'))
        resp = xmlrpclib.loads(resp.body)
        assert resp[0][0] == 'hello world i mean it', resp
        
    def test_system_listmethods(self):
        resp = self.app.post('/xmlrpc', xmlrpclib.dumps((), 'system.listMethods'))
        assert 'addit' in resp, resp
        assert 'genfault' in resp, resp
        assert 'subrpc.joinit' in resp, resp
    
    def test_system_methodsignature(self):
        resp = self.app.post('/xmlrpc', xmlrpclib.dumps(('addit',), 'system.methodSignature'))
        assert 'int' in resp, resp
        assert 'string' in resp, resp
    
    def test_system_methodhelp(self):
        resp = self.app.post('/xmlrpc', xmlrpclib.dumps(('addit',), 'system.methodHelp'))
        assert 'sums an array of numbers' in resp, resp
        resp = self.app.post('/xmlrpc', xmlrpclib.dumps(('subrpc.joinit',), 'system.methodHelp'))
        assert 'joins an array of strings with spaces' in resp, resp
        
    def test_system_methodsignature_bad(self):
        resp = self.app.post('/xmlrpc', xmlrpclib.dumps(('subrpc.nonexistant',), 'system.methodSignature'))
        assert 'Invalid method' in resp, resp

    def test_system_methodhelp_bad(self):
        resp = self.app.post('/xmlrpc', xmlrpclib.dumps(('subrpc.nonexistant',), 'system.methodHelp'))
        assert 'Invalid method' in resp, resp
    
    def test_system_badcall(self):
        resp = self.app.post('/xmlrpc', xmlrpclib.dumps((), 'system.badcall'))
        assert 'Invalid system method called' in resp, resp
        
    def test_xmlrpc_too_many_args(self):
        try:
            resp = self.app.get('/xmlrpc/myurl/what')
        except Exception, e:
            if '404 Not Found' in str(e):
                return
            else:
                raise
        assert 1==0, "too many args passed, and it should not have"

    def test_bad_xmlrpc_call(self):
        resp = self.app.post('/xmlrpc', xmlrpclib.dumps((1,2), 'sumit'))
        assert 'Invalid XMLRPC Method' in resp, resp

    def test_bad_xml_post(self):
        resp = self.app.post('/xmlrpc', 'hello <world "/">')
        assert 'Unable to decode request body' in resp, resp
    
    def test_genfault(self):
        resp = self.app.post('/xmlrpc', xmlrpclib.dumps((1,2), 'genfault'))
        assert '<name>faultCode</name>' in resp, resp
        